/****************************************************************************
 quadrics.cpp
 
 In this demo, a 3D world filled with quadric objects is created for the
 player to explore.
  
 Author   :   Dave Astle
 Date     :   1/5/2001

 Written for OpenGL Game Programming
*****************************************************************************/


/********************************* Includes *********************************/
#define WIN32_LEAN_AND_MEAN   // get rid of Windows things we don't need

#include <windows.h>					// included in all Windows apps
#include <winuser.h>          // Windows constants
#include <gl/gl.h>						// OpenGL include
#include <gl/glu.h>						// OpenGL utilty library
#include <stdlib.h>
#include <math.h>
#include "bitmap.h"
#include <stdio.h>
#include "HiResTimer.h"

#define WND_CLASS_NAME  "OpenGL Window Class"


/*************************** Constants and Macros ***************************/
const int   SCREEN_WIDTH    = 1000;
const int   SCREEN_HEIGHT   = 400;
const int   SCREEN_BPP      = 32;
const bool  USE_FULLSCREEN  = false; 
const char  *APP_TITLE      = "Quadric World";

const float DEFAULT_SPEED   = 0.4f;
const int   WORLD_SIZE      = 250;
// produce a random number between 0.0 and 1.0
#define FRAND (((GLfloat)rand())/RAND_MAX)

#define KEY_DOWN(vk_code)  ((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0)
#define KEY_UP(vk_code)    ((GetAsyncKeyState(vk_code) & 0x8000) ? 0 : 1)


/********************************* Globals **********************************/
HDC       g_hdc;								  // device context
HGLRC     g_hrc;                  // rendering context
BOOL      g_isFullscreen = TRUE;  // toggles fullscreen and windowed display
BOOL      g_isActive = TRUE;      // false if window is minimized
HWND      g_hwnd = NULL;          // main window handle
HINSTANCE g_hInstance;            // application instance

// "player" orientation information
GLdouble  g_playerPos[] = { 0.0, 0.5, 10.0 };
GLdouble  g_lookAt[] = { 0.0, 0.0, 0.0 };
GLfloat   g_viewAngle = -90.0;
GLfloat   g_elevationAngle = 0.0;


// kill this if we can
GLboolean g_keys[256];

// texture names
GLuint  g_fireTexture;
GLuint  g_rocketTexture;
GLuint  g_starsTexture;

// quadric objects
GLUquadricObj *g_normalObject     = NULL;
GLUquadricObj *g_wireframeObject  = NULL;
GLUquadricObj *g_texturedObject   = NULL;
GLUquadricObj *g_flatshadedObject = NULL;


/******************************** Prototypes ********************************/
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

BOOL    SetupWindow(const char *title, int width, int height, int bits, bool isFullscreen);
BOOL    KillWindow();

GLvoid  ResizeScene(GLsizei width, GLsizei height);
BOOL    InitializeScene();
BOOL    DisplayScene();
BOOL    Cleanup();

GLvoid  DrawGround();
GLvoid  DrawStars();
GLvoid  DrawNormalObjects(GLfloat rotation);
GLvoid  DrawWireframeObjects(GLfloat rotation);
GLvoid  DrawTexturedObjects(GLfloat rotation);
GLvoid  DrawFlatshadedObjects(GLfloat rotation);

void    LoadTexture(char *filename, GLuint &texture);


/*****************************************************************************
 WinMain()

 Windows entry point
*****************************************************************************/
int WINAPI WinMain(HINSTANCE g_hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
  MSG   msg;       // message
  BOOL  isDone;    // flag indicatingen the app is done


  // if the window is set up correctly, we can proceed with the message loop
  if (SetupWindow(APP_TITLE, SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, USE_FULLSCREEN))
    isDone = FALSE;
  // otherwise, we need to never enter the loop and proceed to exit
  else
    isDone = TRUE;

  CHiResTimer timer;
  timer.Init();

  int nFrames = 0;

  // main message loop
  while (!isDone)
  {
    if(PeekMessage(&msg, g_hwnd, NULL, NULL, PM_REMOVE))
    {
      if (msg.message == WM_QUIT)		// do we receive a WM_QUIT message?
      {
        isDone = TRUE;				// if so, time to quit the application
      }
      else
      {
        TranslateMessage(&msg);		// translate and dispatch to event queue
        DispatchMessage(&msg);
      }
    }

    // don't update the scene if the app is minimized
    if (g_isActive && g_hrc)
    {
      // update the scene every time through the loop
      DisplayScene();
      // switch the front and back buffers to display the updated scene
      SwapBuffers(g_hdc);
    }
			nFrames++;
			if(nFrames > 100)
				{
				char cOutBuffer[32];

				// Display the fps
				sprintf(cOutBuffer,"%0.1f",timer.GetFPS(100));
				SetWindowText(g_hwnd, cOutBuffer);
				nFrames = 0;
				}
  }

  Cleanup();
  KillWindow();

  return msg.wParam;
} // end WinMain()


/*****************************************************************************
 WndProc()

 Windows message handler
*****************************************************************************/
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch(message)
  {
  case WM_ACTIVATE:  // watch for the window being minimized and restored
    {
      if (!HIWORD(wParam))
      {
        // program was restored or maximized
        g_isActive = TRUE;
      }
      else
      {
        // program was minimized
        g_isActive=FALSE;
      }

      return 0;
    }

  case WM_SYSCOMMAND:  // look for screensavers and powersave mode
    {
      switch (wParam)
			{
				case SC_SCREENSAVE:     // screensaver trying to start
				case SC_MONITORPOWER:   // monitor going to powersave mode
          // returning 0 prevents either from happening
  				return 0;
        default:
          break;
      }
    } break;

  case WM_CLOSE:    // window is being closed
    {
      // send WM_QUIT to message queue
      PostQuitMessage(0);

      return 0;
    }

  case WM_SIZE:
    {
      // update perspective with new width and height
      ResizeScene(LOWORD(lParam), HIWORD(lParam));
      return 0;
    }

  case WM_KEYDOWN:
    {
      switch (toupper(wParam))
      {
      case VK_ESCAPE:
        {
          // send WM_QUIT to message queue
          PostQuitMessage(0);
          return 0;
        }
      default:
        {
          g_keys[wParam] = GL_TRUE;
          return 0;
        }
      };
    }

  case WM_KEYUP:
    {
      g_keys[wParam] = GL_FALSE;
      return 0;
    }
  default:
    break;
  }

  return (DefWindowProc(hwnd, message, wParam, lParam));
} // end WndProc()


/*****************************************************************************
 SetupWindow()

 Create the window and everything else we need, including the device and
 rendering context. If a fullscreen window has been requested but can't be
 created, the user will be prompted to attempt windowed mode. Finally,
 InitializeScene is called for application-specific setup.

 Returns TRUE if everything goes well, or FALSE if an unrecoverable error
 occurs. Note that if this is called twice within a program, KillWindow needs
 to be called before subsequent calls to SetupWindow.
*****************************************************************************/
BOOL SetupWindow(const char *title, int width, int height, int bits, bool isFullscreen)
{
  // set the global flag
  g_isFullscreen = isFullscreen;

  // get our instance handle
  g_hInstance	= GetModuleHandle(NULL);

  WNDCLASSEX  wc;    // window class

  // fill out the window class structure
  wc.cbSize         = sizeof(WNDCLASSEX);
  wc.style          = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
  wc.lpfnWndProc    = WndProc;
  wc.cbClsExtra     = 0;
  wc.cbWndExtra     = 0;
  wc.hInstance      = g_hInstance;
  wc.hIcon          = LoadIcon(NULL, IDI_APPLICATION);	// default icon
  wc.hIconSm        = LoadIcon(NULL, IDI_WINLOGO);		  // windows logo small icon
  wc.hCursor        = LoadCursor(NULL, IDC_ARROW);		  // default arrow
  wc.hbrBackground  = NULL;     // no background needed
  wc.lpszMenuName   = NULL;     // no menu
  wc.lpszClassName  = WND_CLASS_NAME;
  
  // register the windows class
  if (!RegisterClassEx(&wc))
  {
    MessageBox(NULL,"Unable to register the window class", "Error", MB_OK | MB_ICONEXCLAMATION);

    // exit and return FALSE
    return FALSE;       
  }

  // if we're in fullscreen mode, set the display up for it
  if (g_isFullscreen)
  {
    // set up the device mode structure
    DEVMODE screenSettings;
    memset(&screenSettings,0,sizeof(screenSettings));

    screenSettings.dmSize       = sizeof(screenSettings);	
    screenSettings.dmPelsWidth  = width;			// screen width
    screenSettings.dmPelsHeight = height;			// screen height
    screenSettings.dmBitsPerPel = bits;				// bits per pixel
    screenSettings.dmFields     = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

    // attempt to switch to the resolution and bit depth we've selected
    if (ChangeDisplaySettings(&screenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
    {
      // if we can't get fullscreen, let them choose to quit or try windowed mode
      if (MessageBox(NULL, "Cannot run in the fullscreen mode at the selected resolution\n"
                           "on your video card. Try windowed mode instead?",
                           "OpenGL Game Programming",
                           MB_YESNO | MB_ICONEXCLAMATION) == IDYES)
      {
        g_isFullscreen = FALSE;	
      }
      else
			{
				return FALSE;
			}
    }
  }

	DWORD dwExStyle;
	DWORD dwStyle;

  // set the window style appropriately, depending on whether we're in fullscreen mode
	if (g_isFullscreen)
	{
		dwExStyle = WS_EX_APPWINDOW;
		dwStyle = WS_POPUP;						// simple window with no borders or title bar
		ShowCursor(FALSE);            // hide the cursor for now
	}
	else
	{
		dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		dwStyle = WS_OVERLAPPEDWINDOW;
	}

  // set up the window we're rendering to so that the top left corner is at (0,0)
  // and the bottom right corner is (height,width)
  RECT  windowRect;
  windowRect.left = 0;
  windowRect.right = (LONG) width;
  windowRect.top = 0;
  windowRect.bottom = (LONG) height;

  // change the size of the rect to account for borders, etc. set by the style
  AdjustWindowRectEx(&windowRect, dwStyle, FALSE, dwExStyle);

  // class registered, so now create our window
  g_hwnd = CreateWindowEx(dwExStyle,          // extended style
                          WND_CLASS_NAME,     // class name
                          title,              // app name
                          dwStyle |           // window style
                          WS_CLIPCHILDREN |   // required for
                          WS_CLIPSIBLINGS,    // using OpenGL
                          0, 0,               // x,y coordinate
                          windowRect.right - windowRect.left, // width
                          windowRect.bottom - windowRect.top, // height
                          NULL,               // handle to parent
                          NULL,               // handle to menu
                          g_hInstance,        // application instance
                          NULL);              // no extra params

  // see if our window handle is valid
  if (!g_hwnd)
	{
		MessageBox(NULL, "Unable to create window", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

  // get a device context
	if (!(g_hdc = GetDC(g_hwnd)))
	{
		MessageBox(NULL,"Unable to create device context", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

  // set the pixel format we want
  PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),	// size of structure
    1,                              // default version
    PFD_DRAW_TO_WINDOW |            // window drawing support
    PFD_SUPPORT_OPENGL |            // OpenGL support
    PFD_DOUBLEBUFFER,               // double buffering support
    PFD_TYPE_RGBA,                  // RGBA color mode
    bits,                           // 32 bit color mode
    0, 0, 0, 0, 0, 0,               // ignore color bits, non-palettized mode
    0,                              // no alpha buffer
    0,                              // ignore shift bit
    0,                              // no accumulation buffer
    0, 0, 0, 0,                     // ignore accumulation bits
    16,                             // 16 bit z-buffer size
    0,                              // no stencil buffer
    0,                              // no auxiliary buffer
    PFD_MAIN_PLANE,                 // main drawing plane
    0,                              // reserved
    0, 0, 0 };                      // layer masks ignored
      
	GLuint  pixelFormat;

	// choose best matching pixel format
	if (!(pixelFormat = ChoosePixelFormat(g_hdc, &pfd)))
	{
		MessageBox(NULL, "Can't find an appropriate pixel format", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	// set pixel format to device context
  if(!SetPixelFormat(g_hdc, pixelFormat,&pfd))
	{
		MessageBox(NULL, "Unable to set pixel format", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
  }

  // create the OpenGL rendering context
  if (!(g_hrc = wglCreateContext(g_hdc)))
	{
		MessageBox(NULL, "Unable to create OpenGL rendering context", "Error",MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

  // now make the rendering context the active one
  if(!wglMakeCurrent(g_hdc, g_hrc))
	{
		MessageBox(NULL,"Unable to activate OpenGL rendering context", "ERROR", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

  // show the window in the forground, and set the keyboard focus to it
  ShowWindow(g_hwnd, SW_SHOW);
	SetForegroundWindow(g_hwnd);
	SetFocus(g_hwnd);

  // set up the perspective for the current screen size
	ResizeScene(width, height);

  // do one-time initialization
  if (!InitializeScene())
	{
		MessageBox(NULL, "Initialization failed", "Error", MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

  return TRUE;
} // end SetupWindow()


/*****************************************************************************
 KillWindow()

 Deletes the DC, RC, and Window, and restores the original display.
*****************************************************************************/
BOOL KillWindow()
{
  // restore the original display if we're in fullscreen mode
  if (g_isFullscreen)
  {
    ChangeDisplaySettings(NULL, 0);
    ShowCursor(TRUE);
  }

  // if we have an RC, release it
  if (g_hrc)
  {
    // release the RC
    if (!wglMakeCurrent(NULL,NULL))
    {
      MessageBox(NULL, "Unable to release rendering context", "Error", MB_OK | MB_ICONINFORMATION);
    }

    // delete the RC
    if (!wglDeleteContext(g_hrc))
    {
      MessageBox(NULL, "Unable to delete rendering context", "Error", MB_OK | MB_ICONINFORMATION);
    }

    g_hrc = NULL;
  }

  // release the DC if we have one
  if (g_hdc && !ReleaseDC(g_hwnd, g_hdc))
  {
    MessageBox(NULL, "Unable to release device context", "Error", MB_OK | MB_ICONINFORMATION);
    g_hdc = NULL;
  }

  // destroy the window if we have a valid handle
  if (g_hwnd && !DestroyWindow(g_hwnd))
  {
    MessageBox(NULL, "Unable to destroy window", "Error", MB_OK | MB_ICONINFORMATION);
    g_hwnd = NULL;
  }

  // unregister our class so we can create a new one if we need to
  if (!UnregisterClass(WND_CLASS_NAME, g_hInstance))
  {
    MessageBox(NULL, "Unable to unregister window class", "Error", MB_OK | MB_ICONINFORMATION);
    g_hInstance = NULL;
  }

  return TRUE;
} // end KillWindow()


/*****************************************************************************
 ResizeScene()

 Called once when the application starts and again every time the window is
 resized by the user.
*****************************************************************************/
GLvoid ResizeScene(GLsizei width, GLsizei height)
{
  // avoid divide by zero
  if (height==0)  
  {
    height=1;					
  }

  // set the viewport to the new dimensions
  glViewport(0, 0, width, height);

  // select the projection matrix and clear it out
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  // set the perspective with the appropriate aspect ratio
  gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 2500.0f);

  // select modelview matrix
  glMatrixMode(GL_MODELVIEW);
} // end ResizeScene()


/*****************************************************************************
 InitializeScene()

 Performs one-time application-specific setup. Returns FALSE on any failure.
*****************************************************************************/
BOOL InitializeScene()
{
  // enable the zbuffer
  glEnable(GL_DEPTH_TEST);

  // enable lighting
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  // enable using glColor to change material properties
  // we'll use the default glColorMaterial setting (ambient and diffuse)
  glEnable(GL_COLOR_MATERIAL);

  // set the default blending function
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // set up the fog parameters for reflections
  glFogi(GL_FOG_MODE, GL_LINEAR);
  glFogf(GL_FOG_START, -100.0);
  glFogf(GL_FOG_END, 100.0);

  // enable line anti-aliasing and make the lines slightly bigger than default
  glEnable(GL_LINE_SMOOTH);
  glLineWidth(1.5f);

  // create a normal quadric (uses default settings)
  g_normalObject = gluNewQuadric();

  // create an object to use with the wireframe draw stile
  g_wireframeObject = gluNewQuadric();
  gluQuadricDrawStyle(g_wireframeObject, GLU_LINE);

  // created an object that generates texture coordinates
  g_texturedObject = gluNewQuadric();
  gluQuadricTexture(g_texturedObject, GL_TRUE);

  // create an object that uses flat shading
  g_flatshadedObject = gluNewQuadric();
  gluQuadricNormals(g_flatshadedObject, GLU_FLAT);

  // load the textures we need
  LoadTexture("rocket.bmp", g_rocketTexture);
  LoadTexture("fire.bmp", g_fireTexture);
  LoadTexture("stars.bmp", g_starsTexture);

  return TRUE;
} // end InitializeScene()


/*****************************************************************************
 DisplayScene()

 The work of the application is done here. This is called every frame, and
 handles the actual rendering of the scene.
*****************************************************************************/
BOOL DisplayScene()
{
  // see if the player is turning
  if (KEY_DOWN(VK_LEFT))
    g_viewAngle -= 2.0;
  if (KEY_DOWN(VK_RIGHT))
    g_viewAngle += 2.0;

  if (KEY_DOWN('Z'))
    g_elevationAngle += 2.0;
  if (KEY_DOWN('A'))
    g_elevationAngle -= 2.0;

  // calculate the player's angle of rotation in radians
  float rad =  float(3.14159 * g_viewAngle / 180.0f);

  // if the player is pressing up or down, update their x and z coordinates
  // appropriately based on their view angle and speed.
  if (KEY_DOWN(VK_UP))
  {
    g_playerPos[2] += sin(rad) * DEFAULT_SPEED;
    g_playerPos[0] += cos(rad) * DEFAULT_SPEED;
  }
  if (KEY_DOWN(VK_DOWN))
  {
    g_playerPos[2] -= sin(rad) * DEFAULT_SPEED;
    g_playerPos[0] -= cos(rad) * DEFAULT_SPEED;
  }

  // don't allow the player to wander past the "edge of the world"
  if (g_playerPos[0] < -(WORLD_SIZE-50))
    g_playerPos[0] = -(WORLD_SIZE-50);
  if (g_playerPos[0] > (WORLD_SIZE-50))
    g_playerPos[0] = (WORLD_SIZE-50);
  if (g_playerPos[2] < -(WORLD_SIZE-50))
    g_playerPos[2] = -(WORLD_SIZE-50);
  if (g_playerPos[2] > (WORLD_SIZE-50))
    g_playerPos[2] = (WORLD_SIZE-50);


  // use the players view angle to correctly set up the view matrix
  g_lookAt[0] = float(g_playerPos[0] + 100*cos(rad));
  g_lookAt[2] = float(g_playerPos[2] + 100*sin(rad));

  rad = float (3.13149 * g_elevationAngle / 180.0f);

  g_lookAt[1] = float (g_playerPos[1] + 100 * sin(rad));


  // clear the modelview matrix
  glLoadIdentity();

  // setup the view matrix
  gluLookAt(
    g_playerPos[0], g_playerPos[1], g_playerPos[2],
    g_lookAt[0], g_lookAt[1], g_lookAt[2],
    0.0, 1.0, 0.0
    );

  // just clear the depth buffer, no need to clear the color buffer since
  // we write to the whole thing
  glClear(GL_DEPTH_BUFFER_BIT);

  // position the light
  GLfloat pos[4] = { 5.0, 5.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_POSITION, pos);

  // rotation is used for animation
  static GLfloat rotation = 0.0;

  // it's increased by one every frame
  rotation += 1.0;

  // and ranges between 0 and 360
  if (rotation > 360.0)
    rotation = 0.0;

  // draw all of our objects in their normal position
  DrawNormalObjects(rotation);
  DrawWireframeObjects(rotation);
  DrawTexturedObjects(rotation);
  DrawFlatshadedObjects(rotation);

  // draw the sky
  DrawStars();
  
  // enable fog and then draw everything inverted to get
  // a reflection effect
  glPushMatrix();
    glEnable(GL_FOG);

    glScalef(1.0, -1.0, 1.0);

    glLightfv(GL_LIGHT0, GL_POSITION, pos);
    DrawNormalObjects(rotation);
    DrawWireframeObjects(rotation);
    DrawTexturedObjects(rotation);
    DrawFlatshadedObjects(rotation);

    glDisable(GL_FOG);
    DrawStars();
  glPopMatrix();

  // draw the grid representing the ground
  DrawGround();

  return TRUE;
} // end DisplayScene()


/*****************************************************************************
 Cleanup()

 Called at the end of successful program execution.
*****************************************************************************/
BOOL Cleanup()
{
  // delete every valid quadric object
  if (g_normalObject)
    gluDeleteQuadric(g_normalObject);
  if (g_wireframeObject)
    gluDeleteQuadric(g_wireframeObject);
  if (g_texturedObject)
    gluDeleteQuadric(g_texturedObject);
  if (g_flatshadedObject)
    gluDeleteQuadric(g_flatshadedObject);

  return TRUE;
} // end Cleanup()


/*****************************************************************************
 DrawGround()

 Draw a grid of blue lines to represent the ground.
*****************************************************************************/
GLvoid DrawGround()
{
  // enable blending for anti-aliased lines
  glEnable(GL_BLEND);

  // set the color to a bright blue
  glColor3f(0.5f, 0.7f, 1.0f);

  // draw the lines
  glBegin(GL_LINES);
    for (int x = -WORLD_SIZE; x < WORLD_SIZE; x += 6)
    {
      glVertex3i(x, 0, -WORLD_SIZE);
      glVertex3i(x, 0, WORLD_SIZE);
    }

    for (int z = -WORLD_SIZE; z < WORLD_SIZE; z += 6)
    {
      glVertex3i(-WORLD_SIZE, 0, z);
      glVertex3i(WORLD_SIZE, 0, z);
    }
  glEnd();

  // turn blending off
  glDisable(GL_BLEND);
} // end DrawGround()


/*****************************************************************************
 DrawStars()

 Draws five quads textured with star images to represent the sky.
*****************************************************************************/
GLvoid DrawStars()
{
  // enable texturing and choose the stars texture
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, g_starsTexture);

  // we don't care about the quad colors, so just replace them with the texture
  glTexEnvf(GL_TEXTURE_2D, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  glBegin(GL_QUADS);
    // -z plane
    glTexCoord2i(0, 0); glVertex3i(-WORLD_SIZE, 0, -WORLD_SIZE);
    glTexCoord2i(0, 2); glVertex3i(WORLD_SIZE, 0, -WORLD_SIZE);
    glTexCoord2i(1, 2); glVertex3i(WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);
    glTexCoord2i(1, 0); glVertex3i(-WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);

    // x plane
    glTexCoord2i(0, 0); glVertex3i(WORLD_SIZE, 0, -WORLD_SIZE);
    glTexCoord2i(0, 2); glVertex3i(WORLD_SIZE, 0, WORLD_SIZE);
    glTexCoord2i(1, 2); glVertex3i(WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);
    glTexCoord2i(1, 0); glVertex3i(WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);

    // z plane
    glTexCoord2i(0, 0); glVertex3i(WORLD_SIZE, 0, WORLD_SIZE);
    glTexCoord2i(0, 2); glVertex3i(-WORLD_SIZE, 0, WORLD_SIZE);
    glTexCoord2i(1, 2); glVertex3i(-WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);
    glTexCoord2i(1, 0); glVertex3i(WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);

    // -x plane
    glTexCoord2i(0, 0); glVertex3i(-WORLD_SIZE, 0, WORLD_SIZE);
    glTexCoord2i(0, 2); glVertex3i(-WORLD_SIZE, 0, -WORLD_SIZE);
    glTexCoord2i(1, 2); glVertex3i(-WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);
    glTexCoord2i(1, 0); glVertex3i(-WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);

    // y plane, or top
    glTexCoord2i(0, 0); glVertex3i(-WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);
    glTexCoord2i(0, 2); glVertex3i(WORLD_SIZE, WORLD_SIZE, -WORLD_SIZE);
    glTexCoord2i(2, 2); glVertex3i(WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);
    glTexCoord2i(2, 0); glVertex3i(-WORLD_SIZE, WORLD_SIZE, WORLD_SIZE);
  glEnd();

  glDisable(GL_TEXTURE_2D);
} // end DrawStars()


/*****************************************************************************
 DrawNormalObjects()

 Draw several quadrics with the default quadric settings. rotation is used
 for animation purposes, ranges between 0 and 360, and is increased by 1 every
 frame.
*****************************************************************************/
GLvoid DrawNormalObjects(GLfloat rotation)
{
  // make sure the random color values we get are the same every time
  srand(200);

  // save the existing color properties
  glPushAttrib(GL_CURRENT_BIT);

  // a cylinder
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-25.0, 0.0, -4.0);
    glRotatef(-90, 1.0, 0.0, 0.0);
    gluCylinder(g_normalObject, 1.0, 1.0, 3.0, 32, 4);
  glPopMatrix();

  // a tapered cylinder
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(5.0, 0.0, 3.0);
    glRotatef(-90, 1.0, 0.0, 0.0);
    gluCylinder(g_normalObject, 2.0, 0.5, 2.0, 32, 4);
  glPopMatrix();

  // a disk with a hole in it
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(50.0, 1.5, -20.0);
    glRotatef(rotation * 10.0f, 0.0, 1.0, 0.0);
    gluDisk(g_normalObject, 0.7, 1.0, 32, 4);
  glPopMatrix();

  // a cone
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-150.0, 0.5, 0.0);
    glRotatef(-90, 1.0, 0.0, 0.0);
    gluCylinder(g_normalObject, 1.0, 0.0, 2.0, 32, 2);
  glPopMatrix();

  // a bouncing sphere
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);

    // use a sine wave pattern to make the sphere look like it's bouncing
    glTranslatef(5.0f, 3.0f + (float)sin(float(3.14159f*rotation/90.0f)), 50.0f);
    gluSphere(g_normalObject, 2.0f, 32, 20);
  glPopMatrix();

  // restore the previous color values
  glPopAttrib();
} // end DrawNormalObjects()


/*****************************************************************************
 DrawWireframeObjects()

 Draw several objects using the lines, or wireframe, drawing style.
*****************************************************************************/
GLvoid DrawWireframeObjects(GLfloat rotation)
{
  // make sure the random color values we get are the same every time
  srand(300);

  // save the existing color properties
  glPushAttrib(GL_CURRENT_BIT);

  // enable blending to get anti-aliased lines.
  glEnable(GL_BLEND);

  // a couple of spheres that chase each other around
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-20.0f, 2.0f, -20.0f);
    glRotatef(rotation * 2.0f, 1.0f, 0.0f, 0.0f);
    glRotatef(rotation * 2.0f, 0.0f, 1.0f, 0.0f);
    glRotatef(rotation * 2.0f, 0.0f, 0.0f, 1.0f);
    glTranslatef(-0.4f, 0.0f, 0.0f);
    gluSphere(g_wireframeObject, 0.3f, 16, 10);
  glPopMatrix();

  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-20.0f, 2.0f, -20.0f);
    glRotatef(-rotation * 2.0f, 1.0f, 0.0f, 0.0f);
    glRotatef(-rotation * 2.0f, 0.0f, 1.0f, 0.0f);
    glRotatef(-rotation * 2.0f, 0.0f, 0.0f, 1.0f);
    glTranslatef(0.4f, 0.0f, 0.0f);
    gluSphere(g_wireframeObject, 0.3f, 16, 10);
  glPopMatrix();

  // an inverted cone
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-150.0, 0.5, 0.0);
    glRotatef(-90, 1.0, 0.0, 0.0);
    gluCylinder(g_wireframeObject, 0.0, 0.5, 3.0, 32, 4);
  glPopMatrix();

  // a filled disk
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(-40.0, 0.5, 20.0);
    glRotatef(45.0, 1.0, 0.0, 0.0);
    gluDisk(g_wireframeObject, 0.0, 0.5, 32, 4);
  glPopMatrix();

  glDisable(GL_BLEND);

  // restore previous attributes
  glPopAttrib();
} // DrawWireframeObjects()

/*****************************************************************************
 DrawTexturedObjects()

 Draws a few textured objects, including a rocket that moves through the world
*****************************************************************************/
GLvoid DrawTexturedObjects(GLfloat rotation)
{
  // used for the motion of the rocket
  static GLfloat rocketX = 30.0f;
  static GLfloat rocketZ = 30.0f;
  rocketX -= 0.05f;
  rocketZ -= 0.05f;

  if (rocketX < -(GLfloat)WORLD_SIZE)
  {
    rocketX = (GLfloat)WORLD_SIZE;
    rocketZ = (GLfloat)WORLD_SIZE;
  }

  // used in animating the rocket flames
  static GLfloat counter = 0.0f;
  counter += 0.05f;

  if (counter > 1.0f)
    counter = 0.0f;


  // make sure the random color values we get are the same every time
  srand(700);

  // save the existing color properties
  glPushAttrib(GL_CURRENT_BIT);

  // enable texturing and select the rocket texture
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, g_rocketTexture);

  // draw the checked portion of the rocket cylinder
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 128);
    glColor3f(1.0, 1.0, 1.0);
    glTranslatef(rocketX, 3.0, rocketZ);
    glRotatef(-135.0, 0.0, 1.0, 0.0);
    gluCylinder(g_texturedObject, 0.5, 0.5, 3.0, 32, 8);
  glPopMatrix();

  // load the fire texture and draw the rocket's flames
  glBindTexture(GL_TEXTURE_2D, g_fireTexture);
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(1.0, 1.0, 1.0);
    glTranslatef(rocketX, 3.0, rocketZ);
    glRotatef(-135, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 0.0, -1.0);
    glRotatef(rotation, 0.0, 0.0, 1.0);

    // translate the texture coords to move the flames
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
      glTranslatef(0.0, counter, 0.0);
      gluCylinder(g_texturedObject, 0.0, 0.5, 1.5, 64, 8);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  // a rotating sphere
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(100.0, 4.0, -100.0);
    glRotatef(rotation, 0.0, 1.0, 0.0);
    gluSphere(g_texturedObject, 4.0, 32, 20);
  glPopMatrix();

  glDisable(GL_TEXTURE_2D);

  // draw the untextured, white portion of the rocket cylinder
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 128);
    glColor3f(1.0f, 1.0f, 1.0f);
    glTranslatef(rocketX, 3.0f, rocketZ);
    glRotatef(-135.0f, 0.0f, 1.0f, 0.0f);
    glTranslatef(0.0f, 0.0f, 3.0f);
    gluCylinder(g_normalObject, 0.5f, 0.5f, 2.0f, 32, 8);
  glPopMatrix();

  // draw the black nosecone of the rocket
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 128);
    glColor3f(0.2f, 0.2f, 0.2f);
    glTranslatef(rocketX, 3.0f, rocketZ);
    glRotatef(-135, 0.0f, 1.0f, 0.0f);
    glTranslatef(0.0f, 0.0f, 5.0f);
    gluCylinder(g_normalObject, 0.5f, 0.0f, 1.5f, 32, 4);
  glPopMatrix();

  // restore previous attributes
  glPopAttrib();
} // end DrawTexturedObjects()

/*****************************************************************************
 DrawFlatshadedObjects()

 Draws several quadrics using flat shading.
*****************************************************************************/
GLvoid DrawFlatshadedObjects(GLfloat rotation)
{
  glShadeModel(GL_FLAT);
  // used locally for animation
  static GLfloat scaleFactor = 1.0f;
  static GLfloat scaleFactorDelta = 0.001f;

  scaleFactor += scaleFactorDelta;

  if (scaleFactor >= 1.5f || scaleFactor <= 1.0f)
    scaleFactorDelta = -scaleFactorDelta;

  
  // make sure the random color values we get are the same every time
  srand(400);

  // save the existing color properties
  glPushAttrib(GL_CURRENT_BIT);

  // 2 stacked, tapered cylinders
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(83.0f, 0.0f, -10.0f);
    glRotatef(-90, 1.0f, 0.0f, 0.0f);
    glScalef(scaleFactor, scaleFactor, scaleFactor);
    gluCylinder(g_flatshadedObject, 2.0f, 1.0f, 3.0f, 8, 2);
  glPopMatrix();

  glPushMatrix();
    glTranslatef(83.0f, 3.0f * scaleFactor, -10.0f);
    glRotatef(-90, 1.0f, 0.0f, 0.0f);
    glScalef(scaleFactor, scaleFactor, scaleFactor);
    gluCylinder(g_flatshadedObject, 1.0f, 2.0f, 3.0f, 8, 2);
  glPopMatrix();

  // a sphere
  glPushMatrix();
    glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, rand() % 128);
    glColor3f(FRAND, FRAND, FRAND);
    glTranslatef(50.0f, 3.0f, 30.0f);
    glRotatef(-90, 1.0f, 0.0f, 0.0f);
    gluSphere(g_flatshadedObject, 2.0f, 8, 8);
  glPopMatrix();

  // restore previous attributes
  glPopAttrib();
  glShadeModel(GL_SMOOTH);
} // end DrawFlatshadedObjects()


/****************************************************************************
 LoadTexture()

 Loads the texture from the specified file and stores it in iTexture.
*****************************************************************************/
void LoadTexture(char *filename, GLuint &texture)
{
  // get a texture object
  glGenTextures(1, &texture);
  glBindTexture(GL_TEXTURE_2D, texture);

  // load the bitmap
  BITMAPINFOHEADER bitmapInfoHeader;
  unsigned char *buffer = LoadBitmapFile(filename, &bitmapInfoHeader);

  // set up the texture
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, bitmapInfoHeader.biWidth, bitmapInfoHeader.biHeight, 0, GL_RGB, GL_UNSIGNED_BYTE, buffer);

  // we're done with the bitmap data
  free(buffer);
} // end LoadTexture()
